//LIBRARIES

#include "WebSocketsClient.h"
#include <ArduinoJson.h> 
#include <WiFi.h>
#include <Arduino.h>
#include <Ticker.h> 


// PROGRAM EXPLAINED:
// 1) Setup: Run calibration procedure for all emg electrode inputs via "calibration_procedure" function.
//     This procedure measures the average muscle activity when a muscle is flexed and relaxed.
//     Prompts user via serial monitor to flex and unflex the appropriate muscles.
//      - Average of flexed and unflexed is the tolerance for "is_on" function  mentioned later!
//      - Results displayed on serial monitor. Could potentially incorporate LCD display in the future.
//  2) Loop: Consists of the following steps...
//       (1) Read analog values for all electrodes. Record value in "f,w1,w2,e1,e2..." vars
//       (2) Run "is_on" to decide if values read in (1) are in range of average unflexed value. Store in "..._on" vars
//       (3) Update the running arrays of "arr_size" size with "..._on" var for each electrode.
//           Done via "update_array" function for arrays labelled "bin_avg_...". Stores either 1 or 0 for elements.
//       (4) Take average of each array, round up/down to 0 or 1. This determines wether muscle is activated or not
//           Store value in "..._state" vars.

//Websocket
const char* ssid = "Device-Northwestern";
const char* password = "";
const char* IP = "10.106.7.5";  //"10.106.24.41"
WebSocketsClient webSocketEMG;
String send;   
StaticJsonDocument<100> json_read;

//Setup Timers
Ticker timer;
bool read_data = false;
int state_read;
bool done_with_old = true;


// Pins on ESP32 for analog readings
#define EMG_ID 1
#define finger  36 
//#define wrist_1 39 
#define wrist_2 34 
#define elbow_1 35 
#define elbow_2 32 
#define shoulder_1 33 
#define shoulder_2 39

//running binary array size for each electrode
#define arr_size 15  
#define analog_arr_size 50  

// GLOBAL VARIABLES TRACKING STATE OF EMGS
int f_avg, w1_avg, w2_avg, e1_avg, e2_avg, s1_avg, s2_avg;           // Average values (unflexed)
int f_avg_flex, w1_avg_flex, w2_avg_flex, e1_avg_flex, e2_avg_flex, s1_avg_flex, s2_avg_flex; // Average values (flexed)
int f,w1,w2,e1,e2,s1,s2;                                             // current analog values
int f_on,w1_on,w2_on,e1_on,e2_on,s1_on,s2_on;                        // stores if current value outside threshold  
int bin_avg_finger[arr_size],bin_avg_w1[arr_size],bin_avg_w2[arr_size],bin_avg_e1[arr_size],bin_avg_e2[arr_size],bin_avg_s1[arr_size],bin_avg_s2[arr_size];
int f_state,w1_state,w2_state,e1_state,e2_state,s1_state,s2_state;   // Arrays above store previous threshold readings (1 or 0)
                                                                     // "_state" vars store average of those arrays, rounded to 1 or 0
                                                                     // ("_state" represents true state of contraction)
int tol_f,tol_w1,tol_w2,tol_e1,tol_e2,tol_s1,tol_s2;                 // Tolerance: seperation between flexed and unflexed. Mark bounds between flexed and unflexed readings.

// OTHER GLOBAL VARIABLES 
int ticker; 
char msg[100];  // SERIAL MONITOR BUFFER


// FUNCTION DECLARATIONS
void update_running_arrays();
int is_on(int pin_val, int avg, int range);
void calibration_procedure();
  // HELPERS
int avg_arr(int arr[],int samples);
void wait(int delay_sec);
int find_avg(int samples,int pin);
int find_relative_avg(int samples, int pin, int average);
void count_down(int delay_sec);
void easy_calibrate();
int avg_test;


// Only runs if triggered from the outside: there is some event from the outside which triggers 
// the websocketEvent to run.
void webSocketEvent(WStype_t type, uint8_t * payload, size_t length) {
  if (type == WStype_TEXT)
  {
    DeserializationError error = deserializeJson(json_read, payload); 
    if (error) 
    { 
      Serial.print(F("deserializeJson() failed: "));
      Serial.println(error.c_str());
      return;
    } 
    state_read = json_read["EMG_ON"];
  }
}



void readData() {

  if(state_read && done_with_old)
  { 
    done_with_old = false;
     
    // Get current analog values
    //w1 = analogRead(wrist_1);
    f = find_relative_avg(analog_arr_size,finger,f_avg);
    //w1 = find_relative_avg(analog_arr_size,wrist_1,w1_avg);
    w2 = find_relative_avg(analog_arr_size,wrist_2,w2_avg);
    e1 = find_relative_avg(analog_arr_size,elbow_1,e1_avg);
    e2 = find_relative_avg(analog_arr_size,elbow_2,e2_avg);
    s1 = find_relative_avg(analog_arr_size,shoulder_1,s1_avg);
    s2 = find_relative_avg(analog_arr_size,shoulder_2,s2_avg);

    // Determine if current analog measurements are outside thresholds.
    // If within: "_on" = 0, 
    // If outside: "_on" = 1
    f_on =  is_on(f,f_avg,tol_f);
   //w1_on = is_on(w1,w1_avg,tol_w1);
    w2_on = is_on(w2,w2_avg,tol_w2);
    e1_on = is_on(e1,e1_avg,tol_e1);
    e2_on = is_on(e2,e2_avg,tol_e2);
    s1_on = is_on(s1,s1_avg,tol_s1);
    s2_on = is_on(s2,s2_avg,tol_s2);
    // Update runnning binary arrays: 
    //  (these binary arrays store wether previous readings were within or outside the range)
    update_running_arrays();

   // Get current averages of running binary arrays
   // These averages are rounded to integer numbers,
   //  So, average is either "0" (off, muscle not contracting)
   //  or on "1" (muscle is being contracted)
    f_state =  avg_arr(bin_avg_finger,arr_size);
   // w1_state = avg_arr(bin_avg_w1,arr_size);
    w2_state = avg_arr(bin_avg_w2,arr_size);
    e1_state = avg_arr(bin_avg_e1,arr_size);
    e2_state = avg_arr(bin_avg_e2,arr_size);
    s1_state = avg_arr(bin_avg_s1,arr_size);
    s2_state = avg_arr(bin_avg_s2,arr_size);
  
     //sprintf(msg,"BASE: %d, Analog_Wrist1:%d, is_on?:%d, arr_avg:%d ,",w1_avg, w1, w1_on, w1_state);
    //sprintf(msg,"Wrist1:%d, is_on: %d, analog: %d",w1_state,w1_on,w1);
    //sprintf(msg,"Finger:%d,Wrist2: %d, Elbow1: %d, Elbow2: %d, Shoulder1: %d, Shoulder2 %d",f_state,w2_state,e1_state,e2_state,s1_state,s2_state);
    //Serial.println(msg);
    
    // SEND DATA VIA WEBSOCKET
    send = "{\"ID\":";
    send += EMG_ID;
    send += ",\"Finger\":";
    send += f_state;
    //send += ",\"Wrist1\":";
    //send += w1_state;
    send += ",\"Wrist2\":";
    send += w2_state;
    send += ",\"Elbow1\":";
    send += e1_state;
    send += ",\"Elbow2\":";
    send += e2_state;
    send += ",\"Shoulder1\":";
    send += s1_state;
    send += ",\"Shoulder2\":";
    send += s2_state;
    send += "}";
    
    
    Serial.println(send);
    //SEND FINAL STRING
    webSocketEMG.sendTXT(send.c_str(),send.length()); 
    done_with_old = true;  
  }
}



void setup() 
{
  Serial.begin(115200);
  calibration_procedure();
  //easy_calibrate();
  timer.attach_ms(/*rate in secs*/ 40, /*callback*/ readData);

  //WiFi
  WiFi.begin(ssid); 
  while (WiFi.status() != WL_CONNECTED) 
  { 
    Serial.print(".");
    delay(500);
  }
    // Check if connected to wifi
  if(WiFi.status() != WL_CONNECTED) {
    Serial.println("No Wifi!");
    return;
  }
  Serial.println();
  Serial.print("IP Address (AP): "); Serial.println(WiFi.localIP());

  //"10.106.24.41"
  webSocketEMG.begin(IP, 81,"/emg","arduino"); 
  webSocketEMG.onEvent(webSocketEvent);
  webSocketEMG.setReconnectInterval(5000);

  //COMMENT OUT LINE BELOW WHEN USING WEBSOCKET:
 //state_read = 1;
}

void loop() 
{
  webSocketEMG.loop(); 
}




//
// is_on: determines if pin is within threshold or not
// OUTPUT: "1" = outside threshold
//         "0" = within threshold
//
int is_on(int pin_val,int avg, int range)
{
  if(pin_val > range )//+avg || pin_val < avg-range)
  {
    return 1;
  } 
  else
  {
    return 0;
  }
}


//
// Adds new value to binary array of flexed (1) or not flexed (0)
//
void update_running_arrays()
{
  if(ticker == arr_size)
  {
    ticker = 0;
  }
  bin_avg_finger[ticker] = f_on;
  //bin_avg_w1[ticker] = w1_on;
  bin_avg_w2[ticker] = w2_on;
  bin_avg_e1[ticker] = e1_on;
  bin_avg_e2[ticker] = e2_on;
  bin_avg_s1[ticker] = s1_on;
  bin_avg_s2[ticker] = s2_on;
  ticker++;
}

//
//  Helper: takes the average of an input array. 
//  OUTPUT: rounded integer average
//
int avg_arr(int arr[],int samples)
{
  double sum = 0;
  for(int i = 0;i<samples;i++)
  {
    sum += arr[i];
  }
  return round(sum/samples); 
}

void easy_calibrate()
{
  f_avg = 429;
  f_avg_flex = 19;
  tol_f = 10;

  w2_avg = 328;
  w2_avg_flex = 24;
  tol_w2 = 14;

  e1_avg = 284;
  e1_avg_flex = 6;
  tol_e1 = 3;

  e2_avg = 320;
  e2_avg_flex = 7;
  tol_e2 = 4;

  s1_avg = 337;
  s1_avg_flex = 7;
  tol_s1 = 3;

  s2_avg = 429;
  s2_avg_flex = 7;
  tol_s2 = 4;

}

// Run to calibrate all baseline measurements for muscles (flexed and unflexed)
// Tolerance for "is_on" is average of flexed and unflexed: if above average = flexed. Below = unflexed
void calibration_procedure()
{
  //INTRO
  Serial.println("FOR ALL CALIBRATIONS, HOLD OR RELAX MUSCLE AS INSTRUCTED UNTIL TOLD YOU ARE 'DONE'");
  delay(3000);
  Serial.println("HAVE 5 SECONDS TO GET MUSCLE INTO POSITION FOR EACH CALIBRATION");
  delay(4000);
  Serial.println("CALIBRATION PROCEDURES START IN:");
  count_down(3);

  //FINGER
  Serial.println("PLEASE RELAX FINGER.");
  count_down(5);
  f_avg = find_avg(1000,finger);
  Serial.println("DONE");
  delay(1000);
  Serial.println("PLEASE FLEX FINGER.");
  count_down(5);
  f_avg_flex = find_relative_avg(1000,finger,f_avg);
  Serial.println("DONE.");
  delay(1000);
  tol_f = int(f_avg_flex * 0.6);
  Serial.printf("FINGER UNFLEX: %d, FLEX: %d, TOL: %d\n",f_avg,f_avg_flex,tol_f);
  delay(1000);

  
  //WRIST
  /*
  Serial.println("PLEASE RELAX WRIST.");
  count_down(5);
  w1_avg = find_avg(1000,wrist_1);
  Serial.println("DONE.");
  delay(1000);
  Serial.println("PLEASE FLEX WRIST");
  count_down(5);
  w1_avg_flex = find_relative_avg(1000,wrist_1,w1_avg);
  Serial.println("DONE.");
  delay(1000);
  tol_w1 = int(w1_avg_flex * 0.6);
  Serial.printf("W1 UNFLEX: %d, FLEX: %d, TOL: %d\n",w1_avg,w1_avg_flex,tol_w1);
  delay(1000); 
  */
  //PART 2
  Serial.println("PLEASE RELAX WRIST.");
  count_down(5);
  w2_avg = find_avg(1000,wrist_2);
  Serial.println("DONE.");
  delay(1000);
  Serial.println("PLEASE EXTEND WRIST");
  count_down(5);
  w2_avg_flex = find_relative_avg(1000,wrist_2,w2_avg);
  Serial.println("DONE.");
  delay(1000);
  tol_w2 = int(w2_avg_flex*0.6);
  Serial.printf("W2 UNFLEX: %d, FLEX: %d, TOL: %d\n",w2_avg,w2_avg_flex,tol_w2);
  delay(1000);


  // ELBOW
  Serial.println("PLEASE RELAX ELBOW.");
  count_down(5);
  e1_avg = find_avg(1000,elbow_1);
  Serial.println("DONE.");
  delay(1000);
  Serial.println("PLEASE FLEX ELBOW.");
  count_down(5);
  e1_avg_flex = find_relative_avg(1000,elbow_1,e1_avg);
  Serial.println("DONE.");
  delay(1000);
  tol_e1 = int(e1_avg_flex*0.6);
  Serial.printf("E1 UNFLEX: %d, FLEX: %d, TOL: %d\n",e1_avg,e1_avg_flex,tol_e1);
  delay(1000);
  //PART 2
    Serial.println("PLEASE RELAX ELBOW.");
  count_down(5);
  e2_avg = find_avg(1000,elbow_2);
  Serial.println("DONE.");
  delay(1000);
  Serial.println("PLEASE EXTEND ELBOW.");
  count_down(5);
  e2_avg_flex = find_relative_avg(1000,elbow_2,e2_avg);
  Serial.println("DONE.");
  delay(1000);
  tol_e2 = int(e2_avg_flex*0.6);
  Serial.printf("E2 UNFLEX: %d, FLEX: %d, TOL: %d\n",e2_avg,e2_avg_flex,tol_e2);
  delay(1000);


  // SHOULDER
  Serial.println("PLEASE RELAX SHOULDER.");
  count_down(5);
  s1_avg = find_avg(1000,shoulder_1);
  Serial.println("DONE.");
  delay(1000);
  Serial.println("PLEASE ROTATE SHOULDER FORWARD.");
  count_down(5);
  s1_avg_flex = find_relative_avg(1000,shoulder_1,s1_avg);
  Serial.println("DONE.");
  delay(1000);
  tol_s1 = int(s1_avg_flex/2);
  Serial.printf("S1 UNFLEX: %d, FLEX: %d, TOL: %d\n",s1_avg,s1_avg_flex,tol_s1);
  delay(1000);
  //PART 2
  Serial.println("PLEASE RELAX SHOULDER.");
  count_down(5);
  s2_avg = find_avg(1000,shoulder_2);
  Serial.println("DONE.");
  delay(1000);
  Serial.println("PLEASE ROTATE SHOULDER BACKWARDS.");
  count_down(5);
  s2_avg_flex = find_relative_avg(1000,shoulder_2,s2_avg);
  Serial.println("DONE.");
  delay(1000);
  tol_s2 = int(s2_avg_flex*0.6);
  Serial.printf("S2 UNFLEX: %d, FLEX: %d, TOL: %d\n",s2_avg,s2_avg_flex,tol_s2);
  delay(3000);
  
  //FINISH MESSAGE
  Serial.println("CALIBRATION COMPLETE. LOOK AT ABOVE VALUES. IF UNSATISFACTORY, RESTART ESP32 TO REDO PROCESS.");
  //sprintf(msg,"FINGER_FLEX: %d, FINGER_RELAX: %d \n WRIST_FLEX: %d, WRIST_RELAX1: %d \n WRIST_EXTEND: %d");
  delay(4000);
}


// Helper: Prints count down from "delay_sec" to zero
void count_down(int delay_sec)
{
  for(int i = delay_sec;i>0;i--)
  {
    sprintf(msg,"%d",i);
    Serial.println(msg);
    delay(1000);
  }
}


//
// Helper: Finds average value of any input array with "samples" indexes from analog pin "pin"
//         Used for calibration of unflexed values only
int find_avg(int samples,int pin)
{
  int average_array[samples];
  for(int i = 0;i<samples;i++)
  {
    int val = analogRead(pin);
    average_array[i] = val;
  }
  return avg_arr(average_array,samples); 
}


//
// Calculates average for signal based on distance to average. Good for oscillatory signals.
//
int find_relative_avg(int samples, int pin, int average)
{
  int average_array[samples];
  for(int i = 0;i<samples;i++)
  {
    int val = analogRead(pin);
    average_array[i] = abs(val - average);
  }
  return avg_arr(average_array,samples);
}