#include <Arduino.h> 
#include <WebSocketsServer.h>  
#include <ArduinoJson.h>

#include <WiFi.h> 

#include <ASYNCTCP.h>
#include <ESPAsyncWebServer.h>
#include <SPIFFS.h>


//ID system
// + Each client has its own hardcoded ID associated with it.
// + Server on arm receives messages from websockets and decides how to control arm
// + Doesn't actually do any major broadcasting to client: problem with latency avoided
//   Instead, server broadcasts to a website from which a user can connect to see stuff.

#define EMG_ID 1
#define MANUAL_ID 2

//const char* ssid = "DOMEK";
//const char* password = "Snowy123!";
const char* ssid = "Device-Northwestern";
const char* password = "";


// Webserver
AsyncWebServer server(80);
WebSocketsServer webSocketE = WebSocketsServer(81,"/emg","arduino");
WebSocketsServer webSocketM = WebSocketsServer(81,"/manual");//,"arduino");

// Communication
StaticJsonDocument<200> json_in;
String json_out;
char msg[100];

//Define global variables
int emg_id;

int pot_x1;
int pot_x2;
int pot_y1;
int pot_y2;


int adcValue0;
int adcValue1;
int adcValue2;
int adcValue3;

#define Servo0 32   // Elbow
#define Servo1 33   // Finger
#define Servo2 25   // Elbow
#define Servo3 26   // Shoulder

const int PWMFreq = 50;
const int PWMResolution = 10;
const int PWMChannel0 = 0;
const int PWMChannel1 = 1;
const int PWMChannel2 = 2;
const int PWMChannel3 = 3;
// Duty cycle values
int sv0;
int sv1;
int sv2;
int sv3;

int fin;
int wr1;
int wr2;
int el1;
int el2;
int shl1;
int shl2;

bool use_emg,use_man;
String emg_state,man_state;
unsigned old_val,new_val = 0;
unsigned b_avg;


// Function declarations
void update_clients();
void manual_servo_update(int adc_x_1, int adc_y_1, int adc_x_2, int adc_y_2);
void emg_servo_update(int f_fin, int f_wr1, int f_wr2, int f_el1, int f_el2, int f_shl1, int f_shl2);
void update_fin_or_wr(int which_motor, int f_join);
void update_joint(int which_motor, int j_fw, int j_bw);

//
// WEBSERVER RELATED FUNCTIONS
//

// Called in web server handler for various page requests made.
// Sends the right string depending on program parameters.
String processor(const String& var)
{
  if(var == "USE_EMG")
  {
    if(use_emg == 1 && use_man == 0){
      emg_state = "ON";
    }
    else{
      emg_state = "OFF";
    }
    update_clients();
    return emg_state;     //value is sent to webserver to replace USE_EMG in html file
  }
  else if(var == "USE_MANUAL")
  {
    if(use_man == 1 && use_emg == 0){
      man_state = "ON"; 
    }
    else{
      man_state = "OFF";
    }
    update_clients();
    return man_state;     //value is sent to webserver to replace USE_MANUAL in html file
  }
  update_clients();
  return String();
}

//Sends out message to clients telling them wether they should broadcast or not
void update_clients()
{
  json_out = "{\"EMG_ON\":";
  json_out += use_emg;
  json_out += ",\"MAN_ON\":";
  json_out += use_man;
  json_out += "}";
  webSocketE.broadcastTXT(json_out.c_str(),json_out.length());
  webSocketM.broadcastTXT(json_out.c_str(),json_out.length());  
}



//
// WEBSOCKET RELATED FUNCTIONS
//

//
// Whenever anything is sent on any of the EMG websocket type connection, this fucntion is run
//
void webSocketEvent(uint8_t num, WStype_t type, uint8_t *payload, size_t length)
{
  if(type == WStype_TEXT)
  {
      DeserializationError error = deserializeJson(json_in, payload); 
      // Print error msg if incoming String is not JSON formated
      if (error) 
      { 
       Serial.print(F("deserializeJson() failed: "));
       Serial.println(error.c_str());
       return;
      }
      int ID_IN = json_in["ID"];


      if (ID_IN == EMG_ID && use_emg == 1)
      {
        fin = json_in["Finger"];
        wr2 = json_in["Wrist2"];
        el1 = json_in["Elbow1"];
        el2 = json_in["Elbow2"];
        shl1 = json_in["Shoulder1"];
        shl2 = json_in["Shoulder2"];

        sprintf(msg,"(EMG: F: %d, W2: %d, E1: %d, E2: %d, S1: %d, S2: %d)  (Servos: f:%d, w:%d, e:%d, s:%d)",fin,wr2,el1,el2,shl1,shl2,sv0,sv1,sv2,sv3);
        Serial.println(msg);  
        emg_servo_update(fin, wr1, wr2, el1, el2, shl1, shl2);
      }

      else if(ID_IN == MANUAL_ID && use_man == 1)
      {
        pot_x1 = json_in["POT_X1"];
        pot_x2 = json_in["POT_X2"];
        pot_y1 = json_in["POT_Y1"];
        pot_y2 = json_in["POT_Y2"];
        manual_servo_update(pot_x1,pot_y1,pot_x2,pot_y2); 
      }  
  }
  else if (type == WStype_CONNECTED)
  {
    Serial.println("New Client Connected...");
  }
}



//
// SETUP AND LOOP FUNCTIONS
//

void setup() 
{
  Serial.begin(115200);
  //INITIALIZE SPIFFS
  if(!SPIFFS.begin(true)){
  Serial.println("An Error has occurred while mounting SPIFFS");
  return;
  }
  Serial.println("Mounting passes...");  
  
  //CONNECT TO WIFI
  WiFi.begin(ssid);
  while (WiFi.status() != WL_CONNECTED) 
  {
    Serial.print(".");
    delay(500);
  }
  if(WiFi.status() != WL_CONNECTED) {
        Serial.println("No Wifi!");
        return;
  }


   Serial.println();
   Serial.print("IP Address (AP): "); Serial.println(WiFi.localIP());
   // DONE CONNECTING TO WIFI



  // Route for root / web page
  server.on("/", HTTP_GET, [](AsyncWebServerRequest *request){
    Serial.println("GOT HERE = index");
    request->send(SPIFFS, "/index.html", String(), false, processor);
  });
  
  // Route to load style.css file
  server.on("/style.css", HTTP_GET, [](AsyncWebServerRequest *request){
    Serial.println("GOT HERE = style");
    request->send(SPIFFS, "/style.css", "text/css");
  });

  // Route to set EMG to HIGH
  server.on("/on_emg", HTTP_GET, [](AsyncWebServerRequest *request){
    use_emg = true;
    use_man = false;   
    request->send(SPIFFS, "/index.html", String(), false, processor);
  });
  
  // Route to set EMG to LOW
  server.on("/off_emg", HTTP_GET, [](AsyncWebServerRequest *request){
    use_emg = false;    
    request->send(SPIFFS, "/index.html", String(), false, processor);
  });

    // Route to set MAN to HIGH
  server.on("/on_man", HTTP_GET, [](AsyncWebServerRequest *request){
    use_emg = false;
    use_man = true;    
    request->send(SPIFFS, "/index.html", String(), false, processor);
  });
  
  // Route to set MAN to LOW
  server.on("/off_man", HTTP_GET, [](AsyncWebServerRequest *request){
    use_man = false;    
    request->send(SPIFFS, "/index.html", String(), false, processor);
  });
  
  // initialize server and websockets
  server.begin();
  webSocketM.begin();
  webSocketE.begin();
  // handling incoming messages on the websocket
  webSocketM.onEvent(webSocketEvent);
  webSocketE.onEvent(webSocketEvent);


  // Attach Finger
  ledcSetup(PWMChannel0, PWMFreq, PWMResolution);
  ledcAttachPin(Servo0, PWMChannel0);
  // Attach Wrist
  ledcSetup(PWMChannel1, PWMFreq, PWMResolution);
  ledcAttachPin(Servo1, PWMChannel1);
  // Attach Elbow
  ledcSetup(PWMChannel2, PWMFreq, PWMResolution);
  ledcAttachPin(Servo2, PWMChannel2);
  // Attach Shoulder
  ledcSetup(PWMChannel3, PWMFreq, PWMResolution);
  ledcAttachPin(Servo3, PWMChannel3);

  // Set duty cycle values
  sv0 = 60;
  sv1 = 60;
  sv2 = 60;
  sv3 = 60;
  // Set initial pot values
  pot_x1 = 0;
  pot_x2 = 0;
  pot_y1 = 0;
  pot_y2 = 0;

}

void loop() 
{
  webSocketM.loop();
  webSocketE.loop();
}


//
// UPDATE ARM/SERVO FUNCTION
//

void manual_servo_update(int adc_x_1,int adc_y_1, int adc_x_2, int adc_y_2 )
{
  adcValue0 = adc_x_1;
  int dcyc0 = map(adcValue0, 0, 4095, 0, 120);
  bool rng0_g;
  bool rng0_l;
  bool rng0;
  rng0_g = (dcyc0 >= 75);
  rng0_l = (dcyc0 <= 45);
  rng0 = rng0_g || rng0_l;

  adcValue1 = adc_y_1;
  int dcyc1 = map(adcValue1, 0, 4095, 0, 120);
  bool rng1_g;
  bool rng1_l;
  bool rng1;
  rng1_g = (dcyc1 >= 75);
  rng1_l = (dcyc1 <= 45);
  rng1 = rng1_g || rng1_l;

  adcValue2 = adc_x_2;
  int dcyc2 = map(adcValue2, 0, 4095, 0, 120);
  bool rng2_g;
  bool rng2_l;
  bool rng2;
  rng2_g = (dcyc2 >= 75);
  rng2_l = (dcyc2 <= 45);
  rng2 = rng2_g || rng2_l;

  adcValue3 = adc_y_2;
  int dcyc3 = map(adcValue3, 0, 4095, 0, 120);
  bool rng3_g;
  bool rng3_l;
  bool rng3;
  rng3_g = (dcyc3 >= 75);
  rng3_l = (dcyc3 <= 45);
  rng3 = rng3_g || rng3_l;
  
  if (rng0_g)
  {
    if(sv0 < 120)
    {
      sv0++;
    }
    ledcWrite(PWMChannel0, sv0);
  }
  else
  {
    if(rng0_l)
    {
      if(sv0 > 0)
        {
          sv0--;
        }
      ledcWrite(PWMChannel0, sv0);
    }
  }

  if (rng1_g) {
      if(sv1 < 120){
        sv1++;
      }
      ledcWrite(PWMChannel1, sv1);
  }
  else
  {
      if(rng1_l){
        if(sv1 > 0){
          sv1--;
        }
        ledcWrite(PWMChannel1, sv1);
      }
  }

  if (rng2_g) {
    if(sv2 < 120){
      sv2++;
    }
      ledcWrite(PWMChannel2, sv2);
  }
  else
    {
      if(rng2_l){
        if(sv2 > 0){
          sv2--;
        }
        ledcWrite(PWMChannel2, sv2);
      }
  }

  if (rng3_g) {
      if (sv3 < 120){
        sv3++;
      }
      ledcWrite(PWMChannel3, sv3);
  }
  else
  {
      if(rng3_l){
        if(sv3 > 0){
          sv3--;
        }
        ledcWrite(PWMChannel3, sv3);
      }
  } 
  
  // Print duty cycle values of the servo
  sprintf(msg,"MANAUL: (Servos: f:%d, w:%d, e:%d, s:%d)",sv0,sv1,sv2,sv3);
  Serial.println(msg); 
}


//
// EMG
//


void update_fin_or_wr(int which_motor, int dir)
{
  if(which_motor == 0)  //If Finger...
  {
    if (dir == 0 && sv0 < 120)
    {
      sv0 += 1;
    }
    else if(dir == 1 and sv0 > 80)
    {
      sv0 -= 1;
    }
  } 
  else if(which_motor == 1)  //If Wrist...
  {
    if (dir == 0 && sv2 < 80)
    {
      sv2 += 1;
    }
    else if(dir == 1 && sv2 > 50)
    {
      sv2 -= 1;
    }
  }
}



void update_joint(int which_motor, int forward, int backward)
{
  if (which_motor == 2) // If elbow...
  {
    if (forward == 1 && backward == 0 && sv1 < 100)
    {
      sv1 +=1;
    }
    else if(backward == 1 && forward == 0 && sv1 > 40)
    {
      sv1 -=1;
    }
  } 
  else if(which_motor == 3) //If shoulder...
  {
    if (forward == 1 && backward == 0 && sv3 < 80)
    {
      sv3 +=1;
    }
    else if(backward == 1 && forward == 0 && sv3 > 50)
    {
      sv3 -=1;
    }
  }
}



void emg_servo_update(int f_fin, int f_wr1, int f_wr2, int f_el1, int f_el2, int f_shl1, int f_shl2)
{
  // Update duty cycles
  update_fin_or_wr(0, f_fin);
  update_fin_or_wr(1, f_wr2);
  update_joint(2, f_el1, f_el2);
  update_joint(3, f_shl1, f_shl2);
  // Write duty cycles to motors
  ledcWrite(PWMChannel0, sv0);
  ledcWrite(PWMChannel1, sv1);
  ledcWrite(PWMChannel2, sv2);
  ledcWrite(PWMChannel3, sv3);
}


